"""lg-rez / Blocs transversaux

Chaque module de ``lgrez.blocs`` réalise une tâche non liée à une fonctionnalité spécifique du bot : connection à un service, outils divers...

"""

__all__ = ["bdd_tools", "bdd", "env", "gsheets", "pseudoshell", "tools", "webhook"]
